/* NOTE: This file contains sections of code taken from the NVIDIA_CUDA-5.0_Samples
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cuda_runtime.h>

#include "main.h"
#include "GPGPU.h"

// A macro to output the error messages
#define VAR_CHECK_RETURN( value, name )                           \
{                                                                 \
  if ( value < 0 )                                                \
  {                                                               \
    fprintf(stderr, "ERROR: The variable '%s' has not been instantiated. Pleas correct line %d in '%s' by using an API call to get a relevant value.\n", name, -1*value,__FILE__); \
    exit(EXIT_FAILURE);                                           \
  }                                                               \
}

// Defined number of cores for SM of specific compute versions
SMVal coresPerSM[] =
//TODO 8 - replace all 'FIND_REAL_VALUE_IN_REFRENCE' by the correct value source from a reference (try Appendix F of the CUDA C PROGRAMMING GUIDE v5.0)
{
    { 0x10, FIND_REAL_VALUE_IN_REFRENCE },      // Tesla Generation  (SM 1.0) G80   class
    { 0x11, FIND_REAL_VALUE_IN_REFRENCE },      // Tesla Generation  (SM 1.1) G8x   class
    { 0x12, FIND_REAL_VALUE_IN_REFRENCE },      // Tesla Generation  (SM 1.2) G9x   class
    { 0x13, FIND_REAL_VALUE_IN_REFRENCE },      // Tesla Generation  (SM 1.3) GT200 class
    { 0x20, FIND_REAL_VALUE_IN_REFRENCE },     // Fermi Generation  (SM 2.0) GF100 class
    { 0x21, FIND_REAL_VALUE_IN_REFRENCE },     // Fermi Generation  (SM 2.1) GF10x class
    { 0x30, FIND_REAL_VALUE_IN_REFRENCE },    // Kepler Generation (SM 3.0) GK10x class
    { 0x35, FIND_REAL_VALUE_IN_REFRENCE },    // Kepler Generation (SM 3.5) GK11x class
    { 0x37, FIND_REAL_VALUE_IN_REFRENCE },    // Kepler Generation (SM 3.5) GK11x class
    { 0x50, FIND_REAL_VALUE_IN_REFRENCE },    // Kepler Generation (SM 3.5) GK11x class
    { -1, -1 }
};

inline int getValFromSMVer(int major, int minor, SMVal* vals)
{
  int index = 0;

  while (vals[index].SM != -1)
  {
    int thisSM = ((major << 4) + minor);
    int testSM = vals[index].SM;

    if (vals[index].SM == ((major << 4) + minor))
      return vals[index].value;

    index++;
  }
  
  // If we get here we didn't find the value in the array
  return -1;
}

int main(int argc, char **argv)
{
  printf( "\n --== GPGPU workshop CUDA Device Query (Runtime API) ==--\n\n");

  int currentDevvice;
  int deviceCount;
  char pciBusId[1024];
  cudaFuncCache cacheConfig;
  float maxGflops;

  //TODO 1 - Find an API call to calculate the number of devices. Set: deviceCount
  {
    deviceCount = -1*__LINE__;
  }
  VAR_CHECK_RETURN( deviceCount, "deviceCount" );

  if (deviceCount > 0)
  {
    printf("There are %i CUDA enabled devices on this node\n", deviceCount);

    int driverVersion = 0, runtimeVersion = 0;
    //TODO 2 - Find API calls to determine the driver version and runtime version and set driverVersion runtimeVersion
    {
      driverVersion  = -1*__LINE__;
      runtimeVersion = -1*__LINE__;
    }
    VAR_CHECK_RETURN( driverVersion, "driverVersion" );
    VAR_CHECK_RETURN( runtimeVersion, "runtimeVersion" );

    printf("  CUDA Driver Version    %d.%d \n", driverVersion / 1000,
        (driverVersion % 100) / 10);
    printf("  Runtime Version        %d.%d \n", runtimeVersion / 1000,
        (runtimeVersion % 100) / 10);

    for (int device = 0; device < deviceCount; device++)
    {
      cudaDeviceProp deviceProp;

      //TODO 3 - Set the current active device to 'device'
      {
      }

      // Check if the the current device is 'device'
      CUDA_SAFE_CALL( cudaGetDevice(&currentDevvice), "Failed to get device using cudaGetDevice" );
      if ( currentDevvice != device)
      {
        fprintf(stderr, "ERROR: Device not set.\n");
        exit(EXIT_FAILURE);
      }

      //TODO 4 : Use and API call to get the device properties of the current device. Set deviceProp
      {
        deviceProp.clockRate = -1*__LINE__;   // Just seating a random member of deviceProp to give error message.
      }

      // Check if deviceProp has been set
      VAR_CHECK_RETURN( deviceProp.clockRate, "deviceProp" );

      //TODO 5 : Use and API call to get PCI BUS ID of the current device, write result to pciBusId
      {
        sprintf(pciBusId, "Set me %d", __LINE__);
      }

      // Check if pciBusId has been set
      if ( strncmp(pciBusId, "Set me", 6) == 0 )
      {
        int ln = atoi(&(pciBusId[6]));
        fprintf(stderr, "ERROR: The variable 'pciBusId' has not been instantiated. Pleas correct line %d in '%s' by using an API call to get a relevant value.\n", ln, __FILE__); \
        exit(EXIT_FAILURE);
      }

      //TODO 6 : Use and API call to get cache configuration. Set cacheConfig
      {
      }

      printf("\nDevice %d: \"%s\"\n", device, deviceProp.name);
      printf("  CUDA Capability Major.Minor version number:    %d.%d\n",
          deviceProp.major, deviceProp.minor);

      char msg[256];
      sprintf(msg,
          "  Total amount of global memory:                 %.1f GBytes (%llu bytes)\n",
          (float) deviceProp.totalGlobalMem / 1073741824.0f,
          (unsigned long long) deviceProp.totalGlobalMem);
      printf("%s", msg);

      // Check if coresPerSM has been set
      VAR_CHECK_RETURN(getValFromSMVer(deviceProp.major, deviceProp.minor, coresPerSM), "coresPerSM" )

      printf(
          "  (%2d) Multiprocessors x (%3d) CUDA Cores/MP:    %d CUDA Cores\n",
          deviceProp.multiProcessorCount,
          getValFromSMVer(deviceProp.major, deviceProp.minor, coresPerSM),
          getValFromSMVer(deviceProp.major, deviceProp.minor, coresPerSM)
              * deviceProp.multiProcessorCount);
      printf(
          "  GPU Clock rate:                                %.0f MHz (%0.2f GHz)\n",
          deviceProp.clockRate * 1e-3f, deviceProp.clockRate * 1e-6f);

      // TODO 7 : Calculate Theoretical max GFLOPS's
      {
        // Well you can easily find this value, try -->  http://www.nvidia.com/object/tesla-servers.html
        // But how is it calculated?
        // This is going to involve some thinking and some research
        // Hint: have a look at Table 2 of the  CUDA C PROGRAMMING ( http://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#arithmetic-instructions )

        maxGflops = -1*__LINE__;
      }

      // Check if maxGflops has been set
      if ( maxGflops < 0 )
      {
        fprintf(stderr, "ERROR: The variable 'maxGflops' has not been calculated. Pleas correct line %d in '%s'. Calculate this value don't hard code it!\n", (int)(-1*maxGflops), __FILE__); \
        exit(EXIT_FAILURE);
      }

      printf("  Max Gigaflops :                                %.1f Gflops\n",
           maxGflops);

      // This is supported in CUDA 5.0 (runtime API device properties)
      printf("  Memory Clock rate:                             %.0f Mhz\n",
          deviceProp.memoryClockRate * 1e-3f);
      printf("  Memory Bus Width:                              %d-bit\n",
          deviceProp.memoryBusWidth);

      if (deviceProp.l2CacheSize)
      {
        printf("  L2 Cache Size:                                 %d bytes (%0.1f KBytes)\n",
            deviceProp.l2CacheSize, deviceProp.l2CacheSize / 1024.0f );
      }

      printf(
          "  Max Texture Dimension Size (x,y,z)             1D=(%d), 2D=(%d,%d), 3D=(%d,%d,%d)\n",
          deviceProp.maxTexture1D, deviceProp.maxTexture2D[0],
          deviceProp.maxTexture2D[1], deviceProp.maxTexture3D[0],
          deviceProp.maxTexture3D[1], deviceProp.maxTexture3D[2]);
      printf(
          "  Max Layered Texture Size (dim) x layers        1D=(%d) x %d, 2D=(%d,%d) x %d\n",
          deviceProp.maxTexture1DLayered[0], deviceProp.maxTexture1DLayered[1],
          deviceProp.maxTexture2DLayered[0], deviceProp.maxTexture2DLayered[1],
          deviceProp.maxTexture2DLayered[2]);

      printf("  Total amount of constant memory:               %lu bytes (%0.1f KBytes)\n",
          deviceProp.totalConstMem, deviceProp.totalConstMem / 1024.0f );
      printf("  Total amount of shared memory per block:       %lu bytes (%0.1f KBytes)\n",
          deviceProp.sharedMemPerBlock, deviceProp.sharedMemPerBlock / 1024.0f );
      printf("  Total number of registers available per block: %d\n",
          deviceProp.regsPerBlock );
      printf("  Warp size:                                     %d\n",
          deviceProp.warpSize);
      printf("  Maximum number of threads per multiprocessor:  %d\n",
          deviceProp.maxThreadsPerMultiProcessor);
      printf("  Maximum number of threads per block:           %d\n",
          deviceProp.maxThreadsPerBlock);
      printf("  Maximum sizes of each dimension of a block:    %d x %d x %d\n",
          deviceProp.maxThreadsDim[0], deviceProp.maxThreadsDim[1],
          deviceProp.maxThreadsDim[2]);
      printf("  Maximum sizes of each dimension of a grid:     %d x %d x %d\n",
          deviceProp.maxGridSize[0], deviceProp.maxGridSize[1],
          deviceProp.maxGridSize[2]);
      printf("  Maximum memory pitch:                          %lu bytes (%0.1f GBytes)\n",
          deviceProp.memPitch, deviceProp.memPitch / 1073741824.0f );
      printf("  Texture alignment:                             %lu bytes\n",
          deviceProp.textureAlignment );
      printf(
          "  Concurrent copy and kernel execution:          %s with %d copy engine(s)\n",
          (deviceProp.deviceOverlap ? "Yes" : "No"),
          deviceProp.asyncEngineCount);
      printf("  Concurrent kernel execution:                   %s\n",
          deviceProp.concurrentKernels ? "Yes" : "No");
      printf("  Run time limit on kernels:                     %s\n",
          deviceProp.kernelExecTimeoutEnabled ? "Yes" : "No");
      printf("  Integrated GPU sharing Host Memory:            %s\n",
          deviceProp.integrated ? "Yes" : "No");
      printf("  Support host page-locked memory mapping:       %s\n",
          deviceProp.canMapHostMemory ? "Yes" : "No");
      printf("  Alignment requirement for Surfaces:            %s\n",
          deviceProp.surfaceAlignment ? "Yes" : "No");
      printf("  Device has ECC support:                        %s\n",
          deviceProp.ECCEnabled ? "Enabled" : "Disabled");
#ifdef WIN32
      printf("  CUDA Device Driver Mode (TCC or WDDM):         %s\n", deviceProp.tccDriver ? "TCC (Tesla Compute Cluster Driver)" : "WDDM (Windows Display Driver Model)");
#endif
      printf("  Device supports Unified Addressing (UVA):      %s\n",
          deviceProp.unifiedAddressing ? "Yes" : "No");
      printf("  PCI Bus:                                       %s\n",
          pciBusId);
      printf("  Device PCI Bus ID / PCI location ID:           %d / %d\n",
          deviceProp.pciBusID, deviceProp.pciDeviceID);

      const char *sComputeMode[] =
          {
              "Default (multiple host threads can use ::cudaSetDevice() with device simultaneously)",
              "Exclusive (only one host thread in one process is able to use ::cudaSetDevice() with this device)",
              "Prohibited (no host thread can use ::cudaSetDevice() with this device)",
              "Exclusive Process (many threads in one process is able to use ::cudaSetDevice() with this device)",
              "Unknown", NULL };

      printf("  Device Cache Configuration:                    ");
      if ( cacheConfig == cudaFuncCachePreferNone)
        printf("Default function cache configuration, no preference\n");
      else if ( cacheConfig == cudaFuncCachePreferShared)
        printf("Prefer larger shared memory and smaller L1 cache\n");
      else if ( cacheConfig == cudaFuncCachePreferShared)
        printf("Prefer larger L1 cache and smaller shared memory\n");
      else if ( cacheConfig == cudaFuncCachePreferEqual)
        printf("Prefer equal size L1 cache and shared memory\n");

      printf("  Compute Mode:\n");
      printf("     < %s >\n", sComputeMode[deviceProp.computeMode]);
    }
  }
  else
  {
    printf("There are no CUDA enabled devices in this node\n");
  }

  return 0;
}
