// mani.h

#ifndef MANI_H_
#define MANI_H_

#define FIX_ME                        -2            // This is just the definition don't do anything here!
#define FIND_REAL_VALUE_IN_REFRENCE   -1*__LINE__   // This is just the definition don't do anything here!

// Defines for GPU Architecture types (using the SM version to determine the # of cores per SM)
typedef struct
{
    int SM; // 0xMm (hexidecimal notation), M = SM Major version, and m = SM minor version  ie 0x12 (18) is compute 1.2
    int value;
} SMVal;

// 32-bit floating-point add, multiply, multiply-add Operations per Clock Cycle per Multiprocessor
// http://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#arithmetic-instructions__throughput-native-arithmetic-instructions
SMVal opsF32_A_M_MAD_perMPperCC[] =
{
    { 0x10, 8 },      // Tesla  Generation (SM 1.0) G80 class
    { 0x11, 8 },      // Tesla  Generation (SM 1.1) G8x class
    { 0x12, 8 },      // Tesla  Generation (SM 1.2) G9x class
    { 0x13, 8 },      // Tesla  Generation (SM 1.3) GT200 class
    { 0x20, 32 },     // Fermi  Generation (SM 2.0) GF100 class
    { 0x21, 48 },     // Fermi  Generation (SM 2.1) GF10x class
    { 0x30, 192 },    // Kepler Generation (SM 3.0) GK10x class
    { 0x35, 192 },    // Kepler Generation (SM 3.5) GK11x class
    { 0x50, 128 },    // Kepler Generation (SM 3.5) GK11x class
    { -1, -1 }
};

// 64-bit floating-point add, multiply, multiply-add Operations per Clock Cycle per Multiprocessor
// http://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#arithmetic-instructions__throughput-native-arithmetic-instructions
SMVal opsF64_A_M_MAD_perMPperCC[] =
{
    { 0x10, 0 },      // Tesla  Generation (SM 1.0) G80 class
    { 0x11, 0 },      // Tesla  Generation (SM 1.1) G8x class
    { 0x12, 0 },      // Tesla  Generation (SM 1.2) G9x class
    { 0x13, 1 },      // Tesla  Generation (SM 1.3) GT200 class
    { 0x20, 16 },     // Fermi  Generation (SM 2.0) GF100 class
    { 0x21, 4 },      // Fermi  Generation (SM 2.1) GF10x class
    { 0x30, 8 },      // Kepler Generation (SM 3.0) GK10x class
    { 0x35, 64 },     // Kepler Generation (SM 3.5) GK11x class
    { 0x50, 1 },     // Kepler Generation (SM 3.5) GK11x class
    { -1, -1 }
};



#endif /* MANI_H_ */
