/*
 * arguments.cpp
 *
 *  Created on: 12 Apr 2013
 *      Author: chris
 */

#include "arguments.h"

#include <stdio.h>
#include <string.h>

int readTextVal ( char* line, char* name, void* value, int type )
{
  int len1 = strlen(name);
  
  char* next = line+len1;
  if ( strncmp ( line, name, len1 ) == 0 && (*next == ' ' || *next == '\t') )
  {
    line += len1;
    
    if ( line )
    {
      if      ( type == ARG_INT )
      {
        if ( sscanf ( line,  "%i", (int*)value ) != 1 )
        {
          return RD_FORMATERROR; // Format Error
        }
        else
          return RD_SUCSESS;
      }
      else if ( type == ARG_UINT )
      {
        if ( sscanf ( line,  "%ui", (unsigned int*)value ) != 1 )
        {
          return RD_FORMATERROR; // Format Error
        }
        else
          return RD_SUCSESS;
      }
      else if ( type == ARG_BOOL )
      {
        // Strip preceding white space
        while ( line[0] == ' ' || line[0] == '\t' )
          line++;

        char first = ((char*)value)[0];

        printf("ERROR: Boolean \n");

        if (first == '1')
        {
          *(bool*)value = true;
          return RD_SUCSESS;
        }
        else if ( first == '0')
        {
          *(bool*)value = true;
          return RD_SUCSESS;
        }

        printf("ERROR: nothing\n");

        return RD_FORMATERROR;

      }
      else if ( type == ARG_FLOAT )
      {
        replaceChar ( line, 'D', 'E' );
        replaceChar ( line, 'd', 'e' );
        if ( sscanf ( line,  "%f", (float*)value ) != 1 )
        {
          return RD_FORMATERROR; // Format Error
        }
        else
          return RD_SUCSESS;
      }
      else if ( type == ARG_DOUBLE )
      {
        replaceChar ( line, 'D', 'E' );
        replaceChar ( line, 'd', 'e' );
        if ( sscanf ( line,  "%lf", (double*)value ) != 1 )
        {
          return RD_FORMATERROR; // Format Error
        }
        else
          return RD_SUCSESS;
      }
      else if ( type == ARG_STRING )
      {
        // Strip preceding white space
        while ( line[0] == ' ' || line[0] == '\t' )
          line++;
        
        sprintf ( ( char* ) value, "%s", line );
        
        // Strip newlines and carriage returns
        line = ( char* )value;
        while ( line[0] )
        {
          if ( line[0] == '\n' || line[0] == '\r' )
          {
            line[0] = 0;
            break;
          }
          line++;
        }
        
        return RD_SUCSESS;
      }
      else 
      {
        return RD_READERROR;
      }
    }
    else
    {
      return RD_FORMATERROR; // Format Error
    }
  }
  return RD_NOMATCH;
}

int readTextList ( char* line, char* name, char* delims, void* values, int type, int* read )
{
  int len1 = strlen(name);

  char* next = line+len1;
  if ( strncmp ( line, name, len1 ) == 0 && (*next == ' ' || *next == '\t') )
  {
    line+=len1;
    
    char tokennn[1024];
    char *token = strtok ( line, delims );
    *read = 0;
    
    if (token)
    {
      while (token)
      {
        strcpy(tokennn, token);
        
        if      ( type == ARG_INT )
        {
          if ( sscanf ( tokennn,  "%i", &(((int*)values)[*read]) ) != 1 )
          {
            return RD_FORMATERROR; // Format Error
          }
        }
        else if ( type == ARG_FLOAT )
        {
          replaceChar ( token, 'D', 'E' );
          replaceChar ( token, 'd', 'e' );
          if ( sscanf ( token,  "%f", &(((float*)values)[*read]) ) != 1 )
          {
            return RD_FORMATERROR; // Format Error
          }
        }
        else if ( type == ARG_FLOAT )
        {
          replaceChar ( tokennn, 'D', 'E' );
          replaceChar ( tokennn, 'd', 'e' );
          if ( sscanf ( tokennn,  "%lf", &(((double*)values)[*read]) ) != 1 )
          {
            return RD_FORMATERROR; // Format Error
          }
        }
        else if ( type == ARG_STRING )
        {
          sprintf ( ( char* ) ((char*)values)+ sizeof(char)*(*read), "%s", tokennn );
        }
        else 
        {
          return RD_READERROR;
        }

        token = strtok ( NULL, delims );
        (*read)++;
      }

      return RD_SUCSESS;
    }
    else
    {
      return RD_FORMATERROR; // Format Error
    }
  }
  return RD_NOMATCH;
}

void readArgValue(const char* string, commandLineArg* arg)
{
  if (arg->type == ARG_INT)
  {
    if (!arg->value)
      arg->value = new int;
    *(int*) arg->value = atoi(string);
  }
  else if (arg->type == ARG_UINT)
  {
    if (!arg->value)
      arg->value = new unsigned int;
    *(unsigned int*) arg->value = strtoul(string, NULL, 0);
  }
  else if (arg->type == ARG_FLOAT)
  {
    if (!arg->value)
      arg->value = new float;
    *(float*) arg->value = atof(string);
  }
  else if (arg->type == ARG_DOUBLE)
  {
    if (!arg->value)
      arg->value = new double;
    *(double*) arg->value = strtod(string, NULL);
  }
  else if (arg->type == ARG_QUAD)
  {
    if (!arg->value)
      arg->value = new long double;
    *(long double*) arg->value = strtold(string, NULL);
  }
  else if (arg->type == ARG_STRING || arg->type == ARG_END)
  {
    if (!arg->value)
      arg->value = new char[1024];
    strcpy((char*) arg->value, string);
  }
  else if (arg->type == ARG_LONG)
  {
    if (!arg->value)
      arg->value = new long;
    *(long*) arg->value = strtol(string, NULL, 0);
  }
  else if (arg->type == ARG_ULONG)
  {
    if (!arg->value)
      arg->value = new unsigned long;
    *(unsigned long*) arg->value = strtoul(string, NULL, 0);
  }
  else if (arg->type == ARG_BOOL)
  {
    bool hasDefault = 1;
    if (!arg->value)
    {
      hasDefault = 0;
      arg->value = new bool;
    }
    if ( string )
    {
      // Convert to lower case string
      char tmpS[1024];
      strcpy(tmpS, string);
      for(int i = 0; i < strlen(string);++i)
        tmpS[i]=tmpS[i]>0x40&&tmpS[i]<0x5b?tmpS[i]|0x60:tmpS[i];

      if ( strcmp(tmpS, "true") == 0 || strcmp(tmpS, "t")  == 0|| strcmp(tmpS, "1" ) == 0  )
        *(bool*) arg->value = 1;
      else if( strcmp(tmpS, "false") == 0 || strcmp(tmpS, "f") == 0 || strcmp(tmpS, "0") == 0  )
        *(bool*) arg->value = 0;
      else
      {
        fprintf(stderr, "WARNING: Read invalid argument '%s' for boolean value.\n", string);
      }
    }
    else if ( hasDefault ) // Flip the value
    {
      *(bool*) arg->value = ! ( *(bool*) arg->value );
    }
    else      // Make true
    {
      *(bool*) arg->value = 1;
    }
  }
  else
  {
    fprintf(stderr, "ERROR: In function %s in file %s, Invalid type\n", __func__, __FILE__ );
  }
};

commandLineArg readNextArg(int* argn, int argc, char **argv, commandLineArg* args)
{
  char* baseArg = argv[*argn];    // The string of the current argument in the list
  (*argn)++;                      // Increment argn, so to point to the next argument in the list

  char  *equals     = NULL;
  bool  hasDash;
  int   argLen      = -1;    	// The length of the text if its a --
  char  cr          = -1;			// the car if single -
  char *rest        = NULL;

  if (baseArg[0] == '-')
  {
    hasDash = true;
    ++baseArg;          // strip first dash

    if (baseArg[0] == '-') // Double dash text
    {
      // Double dash so reading a string command line argument
      ++baseArg;  // strip second dash
    }
    else
    {
      cr = baseArg[0];
    }

    // Find a '=' or set to item in argument list
    equals = strchr(baseArg, '=');

    if ( equals == NULL )
    {
      argLen  = strlen(baseArg);
      if(*argn < argc)
      {
        rest    = argv[*argn];
        if ( rest[0] == '-' )
        {
          // assume this is a flag
          rest = NULL;
        }
        else
        {
          (*argn)++; // Note this "eats" the next parameter if there is an error
        }
      }
    }
    else
    {
      argLen = equals - baseArg;
      rest = equals + 1;
    }
  }
  else // No dashes so just read
  {
    hasDash = false;
    rest = baseArg;
  }

  int idx     = 0;
  int blnks   = 0;

  while (args[idx].type != ARG_END) // loop through input data structure
  {
    if ( args[idx].type != ARG_HELPTEXT ) // Ignore help text lines
    {
      bool argBlank = false;

      if ( args[idx].longName[0] == 0 && args[idx].charName == 0 ) // args list is blank
      {
        if      ( *argn-2 == blnks ) // First blank or blank in 'order'
        {
          argBlank = true;
        }
        else if ( blnks == 0 )
        {
          // First blank in the input so assume its the first blank in the output
          argBlank = true;
        }
        blnks++;
      }

      if ( (argBlank && (argLen==-1) ) || ((argLen == strlen(args[idx].longName)) && !strncmp(baseArg, args[idx].longName, argLen) ) || cr == args[idx].charName )
      {
        if      ( args[idx].type == ARG_NONE     )
        {
          return args[idx];
        }
        else if ( args[idx].type == ARG_PRNTHELP )
        {
          return args[idx];
        }

        if      ( args[idx].type == ARG_TEXTLIST )
        {
          if ( args[idx].value == NULL )
          {
            fprintf(stderr, "ERROR: found a text list item with no value.\n" );
          }
          else
          {
            testList *lst = (testList *)args[idx].value;
            strcpy(lst->list[lst->noEnts], rest );
            lst->noEnts++;

            while ( (*argn) < argc && argv[(*argn)][0] != '-' )
            {
              strcpy(lst->list[lst->noEnts], argv[(*argn)] );
              lst->noEnts++;
              (*argn)++;
            }
            return args[idx];
          }
        }
        else
        {
          // Found the given parameter in the argument list data structure, try read it.
          readArgValue(rest, &(args[idx]));
          return args[idx];
        }
      }
    }
    idx++;
  }

  // spin to end
  idx = 0;
  while (args[idx].type != ARG_END)
    idx++;

  FOLD // Print error message  .
  {
    fprintf(stderr, "WARNING: invalid parameter '%s' found at position %i, its possible value is %s. Ignoring it.\n\n", baseArg, *argn, rest);
    //readArgValue(baseArg, &(args[idx])); // read the actual value form the argument list
  }
  
  return args[idx];
};

void printHelpFromArgs(commandLineArg* args)
{
  int maxLen = 0;
  int idx = 0;
  while (args[idx].type != ARG_END)
  {
    maxLen = strlen(args[idx].longName)+strlen(args[idx].typeDescription)+3 > maxLen ? strlen(args[idx].longName)+strlen(args[idx].typeDescription)+3 : maxLen;
    idx++;
  }

  idx = 0;
  while (args[idx].type != ARG_END)
  {
    if (args[idx].type == ARG_HELPTEXT)
      printf("%s\n", args[idx].description);
    else
    {
      if (args[idx].charName != 0 )
      {
        if (strlen(args[idx].longName))
          printf(" -%c, ", args[idx].charName);
        else
          printf(" -%c  ", args[idx].charName);
      }
      else
        printf  ("     ");

      char tmpStr[1024];

      if (strlen(args[idx].longName))
      {
        sprintf(tmpStr, "--%s", args[idx].longName);
        if (strlen(args[idx].typeDescription))
        {
          sprintf(tmpStr,"%s=%s", tmpStr, args[idx].typeDescription);
        }
      }
      else
        tmpStr[0]=0;

      char fmtstring1[1024];
      sprintf(fmtstring1, " %%-%is ",maxLen);
      printf(fmtstring1,tmpStr);

      printf(" %s\n", args[idx].description);


      if (args[idx].value != NULL )
      {
        sprintf(fmtstring1, "%%-%is",maxLen+8);
        printf(fmtstring1,"");

        if      ( args[idx].type == ARG_INT )
        {
          printf("Default value [%i]\n",*((int*)args[idx].value) );
        }
        else if      ( args[idx].type == ARG_UINT )
        {
          printf("Default value [%ii]\n",*((unsigned int*)args[idx].value) );
        }
        else if      ( args[idx].type == ARG_LONG )
        {
          printf("Default value [%li]\n",*((long*)args[idx].value) );
        }
        else if      ( args[idx].type == ARG_ULONG )
        {
          printf("Default value [%li]\n",*((unsigned long*)args[idx].value) );
        }
        else if ( args[idx].type == ARG_FLOAT )
        {
          printf("Default value [%f]\n",*((float*)args[idx].value) );
        }
        else if ( args[idx].type == ARG_DOUBLE )
        {
          printf("Default value [%f]\n",*((double*)args[idx].value) );
        }
        else if ( args[idx].type == ARG_STRING )
        {
          printf("Default value [%s]\n", (char*)args[idx].value );
        }
        else if      ( args[idx].type == ARG_BOOL )
        {
          if ( *((bool*)args[idx].value) == true)
            printf("Default value [True]\n" );
          else
            printf("Default value [False]\n" );
        }
        else
        {
          printf("Default value [ I don't know what data type this is!]\n");
        }

      }
    }

    idx++;
  }
};

int readArgs(int argc, char **argv, commandLineArg* args)
{
  int argN = 1;         // Starts with the first argument 1 ( 0 is the name if the binary)

  while ( argN < argc ) // Loop through arguments
  {
    // read the next command line argument (note this increments argN)
    commandLineArg readArg = readNextArg ( &argN, argc, argv, args );

    if ( readArg.type != ARG_END )
    {
      // Argument was found in the list args, so handle it
      if ( readArg.type == ARG_PRNTHELP )
      {
        printHelpFromArgs ( args );         // Print help message
        exit ( EXIT_SUCCESS );              // Exit
      }
    }
    else                      // Invalid parameter  .
    {
      fprintf ( stderr, "WARNING: invalid parameter '%s' found at position %i. Ignoring it.\n", ( char* ) readArg.value, argN );
      // Invalid parameter
      printHelpFromArgs ( args );         // Print help message
      //exit ( EXIT_FAILURE );              // Exit
    }
  }
}
