/*
 * arguments.h
 *
 *  Created on: 12 Apr 2013
 *      Author: chris
 */

#ifndef ARGUMENTS_H_
#define ARGUMENTS_H_

#include <iostream>
#include <string>
#include <stdio.h>
#include <stdlib.h>

#include "util.h"

#define ARG_NONE        1       ///< Specifying no data type, ie flag only
#define ARG_CHAR        2       ///< Specifying a signed char
#define ARG_UCHAR       3       ///< Specifying a unsigned char
#define ARG_INT         4       ///< Specifying a signed integer
#define ARG_UINT        5       ///< Specifying a unsigned integer
#define ARG_FLOAT       6       ///< Specifying a signed single precision float
#define ARG_DOUBLE      7       ///< Specifying a signed double precision float
#define ARG_QUAD        8       ///< Specifying a signed double precision float
#define ARG_STRING      9       ///< Specifying a string
#define ARG_LONG        10      ///< Specifying a signed long integer
#define ARG_ULONG       11      ///< Specifying a unsigned long integer
#define ARG_BOOL        12      ///< Specifying a boolean value


#define ARG_TEXTLIST    97      ///
#define ARG_PRNTHELP    98      ///
#define ARG_HELPTEXT    99      ///< Specifying a line of text to be printed as the help text
#define ARG_END         100     ///< This indicates the end of a list of command line arguments, as the number or elements in the list is not specified

#define RD_NOMATCH        0
#define RD_SUCSESS        1
#define RD_FORMATERROR   -2
#define RD_READERROR     -3

struct testList
{
    int   noEnts;
    char  list[1024][1024];
};

struct commandLineArg
{
    int     type;                     ///< The of the value, this is specified in this header as ARG_XXXX
    char    charName;                 ///< A char specifying the command line flag, in a argument list this should be proceeded by a single dash "-"
    char    longName[100];            ///< A string specifying the command line argument name, in a argument list this should be proceeded by a double dash "--"
    char    typeDescription[100];     ///< A text string describing the value of the argument, ie: float, int, value, ...
    char    description[1024];        ///< A text string describing the command line argument
    void*   value;                    ///< A pointer to a value of the type specified in type. This will have to be typecast to used!
} ;

/** Print a help message from the args structure 
 * @param args  A list if potential command line arguments. NOTE it must be ended with an argument of the type 'ARG_END'
 */
void printHelpFromArgs(commandLineArg* args);

/** Return the value of the next argument
 *
 * @param argn  The index of the item in the list specified by argv
 * @param argc	The number of parameters in the list argv
 * @param argv  An array of arguments to be parsed
 * @param args  A list of potential command line arguments. NOTE it must be ended with an argument of the type 'ARG_END'
 * @return commandLineArg  The value read, the actual value is in commandLineArg.value * 
 * If the argument at argn is an invalid argument, it will return the last item in the args list (ie, one with the type 'ARG_END' )
 * 
 * argn will be incremented for every item read from argv thus on return it will index the next argument
 */
commandLineArg readNextArg(int* argn, int argc, char **argv, commandLineArg* args);

/** Read all command line arguments
 *
 * @param argc	The number of parameters in tghe list argv
 * @param argv  An array of arguments to be parsed
 * @param args  A list of potential command line arguments. NOTE it must be ended with an argument of the type 'ARG_END'
 * @return
 *
 */
int readArgs(int argc, char **argv, commandLineArg* args);

/** Read a single value associated with a text line
 *
 * @param line    The line to scan
 * @param name    The name of the identifier
 * @param value   A pointer to where the data will be written (must match TYPE)
 * @param type    An identifier of the type
 * @return int    A error code
 **/
int readTextVal ( char* line, char* name, void* value, int type );

/** Read list of values associated with a text line
 *
 * @param line    The line to scan
 * @param name    The name of the identifier
 * @param delims  A list of delimiters
 * @param value   A pointer to where the data will be written (must match TYPE)
 * @param type    An identifier of the type
 * @param read
 * @return int    A error code
 **/
int readTextList ( char* line, char* name, char* delims, void* values, int type, int* read );

#endif /* ARGUMENTS_H_ */
