#include "arrayOps.h"

void randArray  ( float* array, const unsigned long noEls, const float min, const float max )
{
  // Generate random numbers in parallel on the CPU
  
  #pragma omp parallel
  {
    // NB: A different seed for each thread
    unsigned int seed = time ( NULL ) ^ omp_get_thread_num();
   
    #pragma omp for
    for ( unsigned long i = 0; i < noEls; i++ )
      array[i] = min + (max-min)*rand_r ( &seed ) / ( float ) RAND_MAX;
  }
}

void randArray ( int* array, unsigned long noEls, int min, int max )
{
  // Generate random numbers in parallel on the CPU

  #pragma omp parallel
  {
    // NB: A different seed for each thread
    unsigned int seed = time ( NULL ) ^ omp_get_thread_num();

    #pragma omp for
    for ( unsigned long long i = 0; i < noEls; i++ )
      array[i] = min + ( int ) ( rand_r ( &seed ) / ( float ) RAND_MAX * max );
  }
}

void randArray ( double* array, unsigned long noEls, double min, double max )
{
  #pragma omp parallel
  {
    // NB: A different seed for each thread
    unsigned int seed = time ( NULL ) ^ omp_get_thread_num();
    
    #pragma omp for
    for ( unsigned long i = 0; i < noEls; i++ )
      array[i] = min + (max-min)*rand_r ( &seed ) / ( double ) RAND_MAX;
  }
}

void printArrayiFillLine ( int noEls, int len )
{
  printf ( "+" );

  for ( int e = 0; e < noEls; e++ )
  {
    for ( int l = 0; l < len; l++ )
      printf ( "-" );

    printf ( "+" );
  }
}

void printArrayiLinei ( int *arrayA, int noEls, int len )
{
  char frmt[1024];
  sprintf ( frmt, "%%%id|", len );

  printf ( "|" );

  for ( int e = 0; e < noEls; e++ )
  {
    printf ( frmt, arrayA[e] );
  }
}

void printArrayiLinef ( float *arrayA, int noEls, int len )
{
  char frmt[1024];
  sprintf ( frmt, "%%%i.2f|", len );

  printf ( "|" );

  for ( int e = 0; e < noEls; e++ )
  {
    printf ( frmt, arrayA[e] );
  }
}

void printArray ( int *arrayA, int *arrayB, int *arrayC, unsigned int m, unsigned int n, unsigned int o )
{
  int strLen = 3;

  char frmt[1024];
  sprintf ( frmt, "%%%is", n * ( strLen + 1 ) + 3 );


  printf ( frmt, "" );
  printArrayiFillLine ( o, strLen );
  printf ( "\n" );

  for ( int nn = 0; nn < n; nn++ )
  {
    printf ( frmt, "" );
    printArrayiLinei ( &arrayB[nn * o], o, strLen );
    printf ( "\n" );

    printf ( frmt, "" );
    printArrayiFillLine ( o, strLen );
    printf ( "\n" );
  }

  printf ( "\n" );

  printArrayiFillLine ( n, strLen );
  printf ( "  " );
  printArrayiFillLine ( o, strLen );
  printf ( "\n" );

  for ( int nn = 0; nn < m; nn++ )
  {
    printArrayiLinei ( &arrayA[nn * n], n, strLen );
    printf ( "  " );
    printArrayiLinei ( &arrayC[nn * o], o, strLen );
    printf ( "\n" );

    printArrayiFillLine ( n, strLen );
    printf ( "  " );
    printArrayiFillLine ( o, strLen );
    printf ( "\n" );
  }
}

void printArray ( float *arrayA, float *arrayB, float *arrayC, unsigned int m, unsigned int n, unsigned int o )
{
  int strLen = 6;

  char frmt[1024];
  sprintf ( frmt, "%%%is", n * ( strLen + 1 ) + 3 );


  printf ( frmt, "" );
  printArrayiFillLine ( o, strLen );
  printf ( "\n" );

  // Print the second array on top
  for ( int nn = 0; nn < n; nn++ )
  {
    printf ( frmt, "" );
    printArrayiLinef ( &arrayB[nn * o], o, strLen );
    printf ( "\n" );

    printf ( frmt, "" );
    printArrayiFillLine ( o, strLen );
    printf ( "\n" );
  }

  printf ( "\n" );

  // Print a blank line
  printArrayiFillLine ( n, strLen );
  printf ( "  " );
  printArrayiFillLine ( o, strLen );
  printf ( "\n" );

  // Print first and third matrix line by line
  for ( int nn = 0; nn < m; nn++ )
  {
    // Print line of first matrix
    printArrayiLinef ( &arrayA[nn * n], n, strLen );
    printf ( "  " );
    printArrayiLinef ( &arrayC[nn * o], o, strLen );
    printf ( "\n" );

    // Print line of third matrix
    printArrayiFillLine ( n, strLen );
    printf ( "  " );
    printArrayiFillLine ( o, strLen );
    printf ( "\n" );
  }
}

float absDiff ( float *arrayA, float *arrayB, unsigned long n )
{
  float absDiff = 0;

  #pragma omp parallel for reduction(+:absDiff)         // Do the check in parallel on the CPU
  for ( unsigned long i = 0; i < n; i++ )
  {
    absDiff += fabs ( arrayA[i] - arrayB[i] );
  }

  return absDiff;
}

float maxAbsDiff ( float *arrayA, float *arrayB, unsigned long n )
{

  float *maxDiffArr = ( float* ) malloc ( omp_get_max_threads() * sizeof ( float ) );

  for ( int i = 0; i < omp_get_max_threads(); i++ )
    maxDiffArr[i] = 0.0f;

  #pragma omp parallel
  {
    int thread = omp_get_thread_num();
    float absDiff;

    #pragma omp for
    for ( unsigned long i = 0; i < n; i++ )
    {
      absDiff = fabs ( arrayA[i] - arrayB[i] );
      maxDiffArr[thread] = absDiff > maxDiffArr[thread] ? absDiff : maxDiffArr[thread];
    }
  }

  float maxDiff = 0.0;

  for ( int i = 0; i < omp_get_max_threads(); i++ )
  {
    maxDiff = maxDiff > maxDiffArr[i] ? maxDiff : maxDiffArr[i];
  }

  free(maxDiffArr);
  return maxDiff;
}

int absDiff ( int *arrayA, int *arrayB, unsigned long n )
{
  int absDiff = 0;

  #pragma omp parallel for reduction(+:absDiff)         // Do the check in parallel on the CPU
  for ( unsigned long i = 0; i < n; i++ )
  {
    absDiff += abs ( arrayA[i] - arrayB[i] );
  }

  return absDiff;
}

int maxAbsDiff ( int *arrayA, int *arrayB, unsigned long n )
{

  int *maxDiffArr = ( int* ) malloc ( omp_get_max_threads() * sizeof ( int ) );

  for ( int i = 0; i < omp_get_max_threads(); i++ )
    maxDiffArr[i] = 0;

  #pragma omp parallel
  {
    int thread = omp_get_thread_num();
    int absDiff;

    #pragma omp for
    for ( unsigned long i = 0; i < n; i++ )
    {
      
      //absDiff = abs ( arrayA[i] - arrayB[i] );
      //maxDiffArr[thread] = absDiff > maxDiffArr[thread] ? absDiff : maxDiffArr[thread];
      int val = arrayA[i] - arrayB[i];
      int absDiff2 = MAX(val,-val);
      maxDiffArr[thread] = absDiff2 > maxDiffArr[thread] ? absDiff2 : maxDiffArr[thread];
    }
  }
  
  int maxDiff = 0;
  for ( int i = 0; i < omp_get_max_threads(); i++ )
  {
    maxDiff = maxDiff > maxDiffArr[i] ? maxDiff : maxDiffArr[i];
  }
  
  free(maxDiffArr);
  return maxDiff;
}

void normalize(float *array, unsigned long n)
{
  basicStats details = calcDetails(array, n);
  
  #pragma omp parallel for
  for ( unsigned long i = 0; i < n; i++ )
  {
    array[i] = (array[i]-details.mean)/(details.sigma);
  }  
}

