#include <iostream>
#include <fstream>
#include <time.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include <cstdio>

#include "log.h"

void printTextLine ( int lineWidth, char* text, char edge, char fill )
{
  printf ( "%c", edge );

  int len = strlen ( text );

  int bf = ( lineWidth - len )  / 2;

  for ( int i = 0 ; i < bf; i++ )
    printf ( "%c", fill );

  if ( ( int ) ( lineWidth - len ) % 2 )
    printf ( "%c", fill );

  printf ( "%s", text );

  for ( int i = 0 ; i < bf; i++ )
    printf ( "%c", fill );

  printf ( "%c\n", edge );
}

void printTextLine2 ( int before, int after, int lineWidth, char edge, char fill, const char* text, ... )
{
  for ( int i = 0; i < before; i++ )
    printf ( "\n" );

  printf ( "%c", edge );

  char txt[lineWidth];
  
  // write variable
  va_list ap;
  va_start ( ap, text );
  sprintf ( txt, text, ap );
  va_end ( ap );
  

  int len = strlen ( txt );

  int bf = ( lineWidth - len )  / 2;

  for ( int i = 0 ; i < bf; i++ )
    printf ( "%c", fill );

  if ( ( int ) ( lineWidth - len ) % 2 )
    printf ( "%c", fill );

  printf ( "%s", txt );


  for ( int i = 0 ; i < bf; i++ )
    printf ( "%c", fill );

  printf ( "%c\n", edge );

  for ( int i = 0; i < after; i++ )
    printf ( "\n" );
}

Logger::Logger()
{
  init();
}

Logger::Logger ( char* filename, bool append )
{
  init();
  setFile ( filename, append);
}

void Logger::init()
{
  myfile = NULL;

  lineNo = 0;
  delim = ';';
  cvVarNo = 0;

  newList = false;

  memset ( headLine1, 0, sizeof ( headLine1 ) );
  memset ( headLine2, 0, sizeof ( headLine2 ) );
  memset ( valLine, 0, sizeof ( valLine ) );

  tPointer = headLine1;
  vPointer = valLine;
  tPointer2 = headLine2;

  strClear( vPointer );
  strClear( tPointer );
  strClear( tPointer2 );
  strClear( fname );
}

void Logger::setFile ( char* filename, bool append  )
{
  if ( myfile != NULL )
    fclose ( myfile );

  if (append )
    myfile = fopen ( filename, "a+" );
  else
    myfile = fopen ( filename, "w" );
  
  if ( myfile == NULL )
    fprintf ( stderr, "ERROR: Unable to open log file %s\n", filename );
  else
  {
    sprintf(fname,"%s",filename);
    fseek(myfile, 0, SEEK_SET);
    char* n1 = fgets ( headLine1, sizeof(headLine1), myfile );
    if(!ferror(myfile))
    {
      if ( strncmp(headLine1, "#;", 2) == 0 )
      {
        sprintf(headLine1,"%s", tPointer+2);

        // Now try reed the second line
        char* n2 = fgets ( headLine2, sizeof(headLine2), myfile );
        if(!ferror(myfile))
        {
          if ( strncmp(headLine2, "#;", 2) == 0 )
          {
            sprintf(headLine2,"%s", tPointer2+2);
          }
          else
          {
            // Assume 2 is the same as one
            sprintf(headLine2,"%s", tPointer);
          }
        }
        else
        {
          // Assume 2 is the same as one
          sprintf(headLine2,"%s", tPointer);
        }
      }
      else
      {
        strClear(headLine1);
      }
    }
    else
    {
      strClear(headLine1);
    }

    fseek(myfile, 0, SEEK_END);
    //= fgetc(myfile);
  }
}

void Logger::TimeWrite ( char* line )
{
  if ( myfile != NULL )
  {
    // get the current time
    time_t rawtime;
    tm * ptm;
    time ( &rawtime );
    ptm = gmtime ( &rawtime );

    // Print line with time stamp
    fprintf ( myfile, "%04i/%02i/%02i %02i:%02i:%02i | %s",
              1900 + ptm->tm_year,
              ptm->tm_mon,
              ptm->tm_mday,
              ptm->tm_hour,
              ptm->tm_min,
              ptm->tm_sec,
              line
            );
  }
}

void Logger::write ( char* line )
{
  if ( myfile != NULL )
    fprintf ( myfile, "%s", line );
}

void Logger::write ( const char* Format, ... )
{
  if ( myfile != NULL )
  {
    va_list ap;
    va_start ( ap, Format );
    vfprintf ( myfile, Format, ap );
    va_end ( ap );
  }
}

void Logger::lineWrite ( char* line )
{
  if ( myfile != NULL )
  {
    fprintf ( myfile, "%i;%s", lineNo, line );  // Write the line
    lineNo++;                                   // Increment line number
  }
}

void Logger::lineWrite ( const char* Format, ... )
{
  if ( myfile != NULL )
  {
    va_list ap;
    va_start ( ap, Format );
    vfprintf ( myfile, Format, ap );  // Write the line
    va_end ( ap );
    lineNo++;                         // Increment line number
  }
}

void Logger::csvWrite ( const char* Title, const char* Format, ... )
{
  if ( myfile != NULL )
  {
    // The length of the title passed in
    int tlen = strlen ( Title );
    
    // The number of characters written to the title line
    int numWritten;

    // NOTE: Should really check for delim and use that
    if ( strncmp ( Title, tPointer, tlen ) == 0 )
    {
      // The heading matches the current column header, so check 2nd heading
      
      tPointer += tlen + 1;         // Sift past the 1st title

      if ( strncmp ( Title, tPointer2, tlen ) == 0 )
      {
        tPointer2 += tlen + 1;      // Sift past the 1st title
      }
      else
      {
        fprintf ( stderr, "ERROR: Logger 2nd heading mismatch, the cvs may be a bit funny.\n" );

        // Add to header!
        numWritten = sprintf ( tPointer2, "%s%c", Title, delim );

        if ( numWritten >= 0 )
          tPointer2 += numWritten;
      }
    }
    else
    {
      // Add to 1st header!
      numWritten = sprintf ( tPointer, "%s%c", Title, delim );

      if ( numWritten >= 0 )
      {
        newList = true;
        tPointer += numWritten;
      }

      // Add to 2nd header!
      numWritten = sprintf ( tPointer2, "%s%c", Title, delim );

      if ( numWritten >= 0 )
      {
        tPointer2 += numWritten;
      }
    }

    {
      // write variable
      va_list ap;
      va_start ( ap, Format );
      numWritten = vsprintf ( vPointer, Format, ap );   // Write the variable
      va_end ( ap );            
      if ( numWritten >= 0 )
        vPointer += numWritten;                         // Increment the variable pointer

      numWritten = sprintf ( vPointer, "%c", delim );   // Add a deliminator      
      if ( numWritten >= 0 )
        vPointer += numWritten;                         // Increment the variable pointer
    }
    cvVarNo++;                                          // Increment line number
  }
}

void Logger::csvWrite ( const char* Title, const char* Title2, const char* Format, ... )
{
  if ( myfile != NULL )
  {
    // The length of the 1st title passed in
    int tlen = strlen ( Title );
    // The length of the 2st title passed in
    int tlen2 = strlen ( Title2 );

    // The number of characters written to the title line
    int numWritten;

    if ( strncmp ( Title, tPointer, tlen ) == 0 )
    {
      // The heading matches the current column header, so check 2nd heading

      if ( strncmp ( Title2, tPointer2, tlen2 ) == 0 )
      {
        tPointer2 += tlen2 + 1;
      }
      else
      {
        // Add to header!
        numWritten = sprintf ( tPointer2, "%s%c", Title2, delim );

        if ( numWritten >= 0 )
        {
          newList = true;
          tPointer2 += numWritten;
        }
      }

      tPointer += tlen + 1;
    }
    else
    {
      // Add to header!
      numWritten = sprintf ( tPointer, "%s%c", Title, delim );

      if ( numWritten >= 0 )
      {
        newList = true;
        tPointer += numWritten;
      }

      // Add to header!
      numWritten = sprintf ( tPointer2, "%s%c", Title2, delim );

      if ( numWritten >= 0 )
        tPointer2 += numWritten;
    }

    // write variable
    {
      va_list ap;
      va_start ( ap, Format );
      numWritten = vsprintf ( vPointer, Format, ap );
      va_end ( ap );

      if ( numWritten >= 0 )
      {
        vPointer += numWritten;
      }

      numWritten = sprintf ( vPointer, "%c", delim );

      if ( numWritten >= 0 )
      {
        vPointer += numWritten;
      }
    }

    cvVarNo++;
  }

}

void Logger::csvBlankLine()
{
  fprintf ( myfile, "%i%c\n", lineNo, delim );
  lineNo++;
}

void Logger::csvEndLine()
{
  if ( myfile != NULL )
  {
    if ( newList )
    {
      // Print the headings
      fprintf ( myfile, "#%c%s\n", delim, headLine1 );

      // Only print the second heading if it differs from the first
      if ( strcmp(headLine1, headLine2) != 0 )
        fprintf ( myfile, "#%c%s\n", delim, headLine2 );

      newList = false;
    }

    // Print the var line
    fprintf ( myfile, "%i%c%s\n", lineNo, delim, valLine );
    lineNo++;                       // Starts at 0
    tPointer = headLine1;
    vPointer = valLine;
    tPointer2 = headLine2;

    // write to the file
    fflush ( myfile );
  }
}

void Logger::close()
{
  if ( myfile != NULL )
  {
    long size=ftell (myfile);
    fclose ( myfile );
    myfile = NULL;

    if( size == 0 ) // Lets not write blank files
    {
      std::remove(&fname[0]);
    }
  }
}

Logger::~Logger()
{
  if ( myfile != NULL )
    fclose ( myfile );
}

void* spin(void *ptr )
{
  spiner* sp = (spiner *) ptr;

  printf("\r%s                                                ", sp->msg );
  
  int ite = 0;

  while ( sp->msg[0] != 0 )
  {
    ite++;
    
    int no = 2;

    if      ( no == 0 )
    {
      if      (ite == 1 || ite == 5 || ite == 9 )
        printf("\r%s|   ", sp->msg);
      else if (ite == 2 || ite == 6)
        printf("\r%s╱   ", sp->msg);
      else if (ite == 3 || ite == 7)
        printf("\r%s―   ", sp->msg);
      else if (ite == 4 || ite == 8)
        printf("\r%s╲   ", sp->msg);
      if ( ite == 8)
      {
        ite = 0;
      }
    }
    else if ( no == 1 )
    {
      if      (ite == 1 )
        printf("\r%s▖   ", sp->msg);
      if      (ite == 2 )
        printf("\r%s▘   ", sp->msg);
      if      (ite == 3 )
        printf("\r%s▝   ", sp->msg);
      if      (ite == 4 )
      {
        printf("\r%s▗   ", sp->msg);
        ite = 0;
      }
    }
    else if ( no == 2 )
    {
      if      (ite == 1 )
        printf("\r%s⌜   ", sp->msg);
      if      (ite == 2 )
        printf("\r%s⌝   ", sp->msg);
      if      (ite == 3 )
        printf("\r%s⌟   ", sp->msg);
      if      (ite == 4 )
      {
        printf("\r%s⌞   ", sp->msg);
        ite = 0;
      }
    }
    else if ( no == 3 )
    {
      if      (ite == 1 )
        printf("\r%s◜   ", sp->msg);
      if      (ite == 2 )
        printf("\r%s◝   ", sp->msg);
      if      (ite == 3 )
        printf("\r%s◞   ", sp->msg);
      if      (ite == 4 )
      {
        printf("\r%s◟   ", sp->msg);
        ite = 0;
      }
    }
    else if ( no == 4 )
    {
      if      (ite == 1 )
        printf("\r%s↑  ", sp->msg);
      if      (ite == 2 )
        printf("\r%s↗  ", sp->msg);
      if      (ite == 3 )
        printf("\r%s→  ", sp->msg);
      if      (ite == 4 )
        printf("\r%s↘  ", sp->msg);
      if      (ite == 5 )
        printf("\r%s↓  ", sp->msg);
      if      (ite == 6 )
        printf("\r%s↙  ", sp->msg);
      if      (ite == 7 )
        printf("\r%s←  ", sp->msg);
      if      (ite == 8 )
      {
        printf("\r%s↖  ", sp->msg);
        ite = 0;
      }
    }

    std::cout.flush();
    usleep(150000);
  }
}

spiner* spinMessage( const char* text, ... )
{
  spiner* sp = new spiner;
  va_list ap;
  va_start ( ap, text );
  int numWritten = vsprintf ( sp->msg, text, ap );
  va_end ( ap );

#ifdef DEBUG
  if ( detect_gdb() )
  {
    //printf("\r%sRunning in GDB so no spin message! ", sp->msg);
    printf("\r%s⟳  ", sp->msg);
    std::cout.flush();
    return 0;
  }
#endif
  pthread_create( &sp->thread, NULL, spin, (void*) sp );
  
  return sp;
}

void killSpinMessage(spiner* sp, int end )
{  
  if (!sp)
    return;

#ifdef DEBUG
	if ( detect_gdb() )
	{
	  if  ( end == SPINEND_NONE )
	  {
	  }
	  else if  ( end == SPINEND_DONE )
	  {
	    printf(" Done  " );
	    std::cout.flush();
	  }
	  else if  ( end == SPINEND_DONENEW )
	  {
	    printf(" Done  \n" );
	    std::cout.flush();
	  }

		return;
	}
#endif
  pthread_cancel(sp->thread);
  int res = pthread_join(sp->thread, NULL);
  std::cout.flush();

  if  ( end == SPINEND_NONE )
  {
  }
  else if  ( end == SPINEND_DONE )
  {
    printf("\r%sDone  ", sp->msg );
  }
  else if  ( end == SPINEND_DONENEW )
  {
    printf("\r%sDone  \n", sp->msg );
  }
  free(sp);
}
