#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <omp.h>

#include <dirent.h>
#include <errno.h>
#include <iostream>
#include <fstream>

#include "util.h"

// Global static pointer used to ensure a single instance of the class.
unitPrinter* unitPrinter::m_pInstance = NULL;

/** This function is called to create an instance of the class.
 *   Calling the constructor publicly is not allowed. The constructor
 *   is private and is only called by this Instance function.
 */
unitPrinter* unitPrinter::Inst()
{
  if ( !m_pInstance ) // Only allow one instance of class to be generated.
    m_pInstance = new unitPrinter;

  return m_pInstance;
}

unitPrinter* unitPrinter::Instance()
{
  if ( !m_pInstance ) // Only allow one instance of class to be generated.
    m_pInstance = new unitPrinter;

  return m_pInstance;
}

char *unitPrinter::timeFromS ( double seconds )
{
  memset ( string, 0, len );
  int n;
  bool min = true;


  if ( seconds < 0.001 )
  {
    sprintf ( string, "0.00 s" );
    return string;
  }

  if ( seconds > 60 * 60 * 24 * 365.25 || !min )
  {
    n = seconds / ( 60 * 60 * 24 * 365.25 );
    sprintf ( string, "%i year", n );

    if ( n > 1 )
      sprintf ( string, "%ss", string );

    seconds -= n * ( 60 * 60 * 24 * 365.25 );
    min = false;
  }

  if ( seconds > 60 * 60 * 24 * 30.4375 || !min )
  {
    n = seconds / ( 60 * 60 * 24 * 30.4375 );

    if ( !min )
      sprintf ( string, "%s ", string );

    sprintf ( string, "%s%i month", string, n );

    if ( n > 1 )
      sprintf ( string, "%ss", string );

    if ( !min )
      return string;

    seconds -= n * ( 60 * 60 ) * 24 * 30.4375;
    min = false;
  }

  if ( seconds > 60 * 60 * 24 || !min )
  {
    n = seconds / ( 60 * 60 * 24 );

    if ( !min )
      sprintf ( string, "%s ", string );

    sprintf ( string, "%s%i day", string, n );

    if ( n > 1 )
      sprintf ( string, "%ss", string );

    if ( !min )
      return string;

    seconds -= n * ( 60 * 60 ) * 24;
    min = false;
  }

  if ( seconds > 60 * 60 || !min )
  {
    n = seconds / ( 60 * 60 );

    if ( !min )
      sprintf ( string, "%s ", string );

    sprintf ( string, "%s%i h", string, n );

    if ( !min )
      return string;

    seconds -= n * ( 60 * 60 );
    min = false;
  }

  if ( seconds > 60 || !min )
  {
    n = seconds / 60;

    if ( !min )
      sprintf ( string, "%s ", string );

    sprintf ( string, "%s%02i m", string, n );

    if ( !min )
      return string;

    seconds -= n * 60;
    min = false;
  }

  if ( seconds > 0 || !min )
  {
    if ( !min )
      sprintf ( string, "%s%02.2f s", string, seconds );
    else
      sprintf ( string, "%s %02.0f s", string, seconds );
  }

  return string;
}

char *unitPrinter::timeFromMS ( double milliSeconds )
{
  int noMS = 1000;

  memset ( string, 0, len );
  int n;
  bool min = true;

  if ( milliSeconds == 0.0 )
  {
    sprintf ( string, "0 ms" );
    return string;
  }

  if ( milliSeconds > 60 * 60 * 24 * 365.25 * noMS  || !min )
  {
    n = milliSeconds / ( 60 * 60 * 24 * 365.25  * noMS );
    sprintf ( string, "%i year", n );

    if ( n > 1 )
      sprintf ( string, "%ss", string );

    milliSeconds -= n * ( 60 * 60 * 24 * 365.25 * noMS );
    min = false;
  }

  if ( milliSeconds > 60 * 60 * 24 * 30.4375 * noMS || !min )
  {
    n = milliSeconds / ( 60 * 60 * 24 * 30.4375 * noMS );

    if ( !min )
      sprintf ( string, "%s ", string );

    sprintf ( string, "%s%i month", string, n );

    if ( n > 1 )
      sprintf ( string, "%ss", string );

    if ( !min )
      return string;

    milliSeconds -= n * ( 60 * 60 ) * 24 * 30.4375 * noMS;
    min = false;
  }

  if ( milliSeconds > 60 * 60 * 24 * noMS || !min )
  {
    n = milliSeconds / ( 60 * 60 * 24 * noMS );

    if ( !min )
      sprintf ( string, "%s ", string );

    sprintf ( string, "%s%i day", string, n );

    if ( n > 1 )
      sprintf ( string, "%ss", string );

    if ( !min )
      return string;

    milliSeconds -= n * ( 60 * 60 ) * 24 * noMS;
    min = false;
  }

  if ( milliSeconds > 60 * 60 * noMS || !min )
  {
    n = milliSeconds / ( 60 * 60 * noMS );

    if ( !min )
      sprintf ( string, "%s ", string );

    sprintf ( string, "%s%i h", string, n );

    if ( !min )
      return string;

    milliSeconds -= n * ( 60 * 60 ) * noMS;
    min = false;
  }

  if ( milliSeconds > 60 * noMS || !min )
  {
    n = milliSeconds / 60 / noMS;

    if ( !min )
      sprintf ( string, "%s ", string );

    sprintf ( string, "%s%02i m", string, n );

    if ( !min )
      return string;

    milliSeconds -= n * 60 * noMS;
    min = false;
  }

  if ( milliSeconds > noMS || !min )
  {
    n = milliSeconds / noMS;

    if ( !min )
      sprintf ( string, "%s ", string );
    else
    {
      sprintf ( string, "%s%06.3f s", string, milliSeconds / (double)noMS );
      return string;
    }

    sprintf ( string, "%s%02i s", string, n );

    if ( !min )
      return string;

    milliSeconds -= n * noMS;
    min = false;
  }

  if ( milliSeconds > 0.0 || !min )
  {
    if ( min )
      sprintf ( string, "%s%05.2f ms", string, milliSeconds );
    else
      sprintf ( string, "%s %03.0f ms", string, milliSeconds );

    min = false;
  }

  return string;
}

char *unitPrinter::sizeFromB ( double byte )
{
  memset ( string, 0, len );
  int n;

  if ( byte > exp10 ( 18 ) )
    sprintf ( string, "%s%.2f EB", string, byte / ( float ) ( exp10 ( 18 ) ) );
  else if ( byte > exp10 ( 15 ) )
    sprintf ( string, "%s%.2f PB", string, byte / ( float ) ( exp10 ( 15 ) ) );
  else if ( byte > exp10 ( 12 ) )
    sprintf ( string, "%s%.2f TB", string, byte / ( float ) ( exp10 ( 12 ) ) );
  else if ( byte > exp10 ( 9 ) )
    sprintf ( string, "%s%.2f GB", string, byte / ( float ) ( exp10 ( 9 ) ) );
  else if ( byte > exp10 ( 6 ) )
    sprintf ( string, "%s%.2f MB", string,  byte / ( float ) ( exp10 ( 6 ) ) );
  else if ( byte > exp10 ( 3 ) )
    sprintf ( string, "%s%.0f kB", string,  byte / ( float ) ( exp10 ( 3 ) ) );
  else
    sprintf ( string, "%s%.0f B", string,  byte );

  return string;
}

double calcDist ( int noDims, float* v1, float* v2 )
{
  double sd = 0;

  for ( int i = 0; i < noDims; i++ )
  {
    sd += ( ( v1[i] - v2[i] ) * ( v1[i] - v2[i] ) );
  }

  if ( sd == 0 )
    return 0;

  return sqrt ( sd );
};

double calcDist ( int noDims, int* v1, int* v2 )
{
  double sd = 0;

  for ( int i = 0; i < noDims; i++ )
  {
    sd += ( ( v1[i] - v2[i] ) * ( v1[i] - v2[i] ) );
  }

  return sqrt ( sd );
};

void getCpuName ( char* name, int core )
{
  char line[MAXLINE];     // a buffer to read data into
  char word1[MAXLINE];     // a buffer to read data into
  char word2[MAXLINE];     // a buffer to read data into
  char word3[MAXLINE];     // a buffer to read data into

  int lineNo = -1;

  sprintf ( name, "No details" );

#if __linux
  FILE * fp;
  fp = popen ( "/bin/cat /proc/cpuinfo | grep '^model name'", "r" );

  if ( fp != NULL )
  {
    while ( !feof ( fp ) )
    {
      if ( fgets ( line, MAXLINE, fp ) )
      {
        lineNo++;

        int cnt =  sscanf ( line,  "model name : %s\n", word1 );

        if ( cnt == 1 )
        {
          if ( lineNo >= core )
          {
            sprintf ( name, "%s", word1 );
            fclose ( fp );
            return;
          }
        }
      }

    }

    fclose ( fp );
  }

  //else
  {
    fprintf ( stderr, "ERROR: Failed to open cpuinfo to get CPU name\n" );
  }

#endif
}

int getNoProc()
{
  int noProc = -1;
#if __linux
  FILE * fp;
  char res[128];
  fp = popen ( "cat /proc/cpuinfo | grep -c '^processor'", "r" );

  if ( fp != NULL )
  {
    size_t read = fread ( res, 1, sizeof ( res ) - 1, fp );
    fclose ( fp );
    sscanf ( res, "%i", &noProc );
  }

#else
  fprintf ( stderr, "WARNING: No way to detect number of cpu's on this system\n" );
#endif
  return noProc;
};

char* splitFileName(char *fName, char* directory, char *file, char *extension)
{
  char temp[1024];
  
  char *lastSlash = strrchr ( fName,  47 );                           // find slash
  char *lastDot   = strrchr ( fName, '.' );                           // find dot
  
  if ( lastSlash )
  {
    if ( directory )
    {
     strncpy(directory, fName, lastSlash-fName );
     directory[lastSlash-fName] = 0;
    }
   
    if ( lastDot )
    {
      if ( file )
      {
       strncpy(file, lastSlash+1, lastDot-lastSlash-1 );
       file[lastDot-lastSlash-1] = 0;
      }
      
      if ( extension )
        sprintf(extension, "%s", lastDot+1);
      else
        strClear(extension);
    }
    else
    {
      if ( file )
        sprintf(file, "%s", lastSlash+1);
    }
    
    return lastSlash+1;
  }
  else
  {
    if (directory)
      strClear(directory);
    
    if ( lastDot )
    {
      if ( file )
      {
        strncpy(file, fName, lastDot-fName );
        file[lastDot-fName]=0;       
      }
      
      if ( extension )
      {
        if (lastDot)
          sprintf(extension, "%s", lastDot+1);
        else
          strClear(extension);
      }
    }
    else
    {
      if ( file )
        sprintf(file, "%s", fName);
    }
    return fName;
  }
}

void replaceChar(char* str, const char replace, const char ch)
{
  char* i = strchr(str, replace);
  
  while (i)
  {
    *i = ch;
    i = strchr(str, replace);
  }
};

int setOmpThreadCound(int *threadCount )
{
  int holdThreads = omp_get_max_threads();
  
  if ( threadCount == NULL )
  {
    omp_set_num_threads(omp_get_max_threads());
  }
  else
  {
    if ( *threadCount <= 0 )
    {
      *threadCount = omp_get_max_threads();
    }
    omp_set_num_threads ( *threadCount );
  }
  return holdThreads;
}

#if    _WIN32
#include <windows.h>
ulong getFreeRam()
{
    MEMORYSTATUSEX status; 
    status.dwLength = sizeof(status);
    GlobalMemoryStatusEx(&status);
    return status.ullTotalPhys;
}
#elif  __linux
#include <sys/sysinfo.h>
ulong getFreeRam()
{
  struct sysinfo sys_info;
  if(sysinfo(&sys_info) != 0)
  {
    fprintf(stderr, "ERROR: Reading memory info.");
    return 0;
  }
  else
    return sys_info.freeram + sys_info.bufferram;
}
#else
ulong getFreeRam()
{
  fprintf(stderr, "ERROR: getFreeRam not enabled on this system.");
}
#endif

#if    _WIN32
#include <windows.h>
size_t getTotalRam()
{
}
#elif  __linux
#include <sys/sysinfo.h>
size_t getTotalRam()
{
  struct sysinfo sys_info;
  if(sysinfo(&sys_info) != 0)
  {
    fprintf(stderr, "ERROR: Reading memory info.");
    return 0;
  }
  else
    return sys_info.totalram;
}
#else
ulong getTotalRam()
{
  fprintf(stderr, "ERROR: getTotalRam not enabled on this system.");
}
#endif

int detect_gdb(void)
{
    int rc = 0;
    FILE *fd = fopen("/tmp", "r");

    if (fileno(fd) >= 5)
    {
        rc = 1;
    }

    fclose(fd);
    return rc;
}

void* safemalloc (ulong size)
{
  double per = getFreeRam() + size / ( (double)getTotalRam() );
  if ( per > 0.98 )
  {
    fprintf(stderr,"ERROR in %s there is insufficient memory to allocate %s of memory.\n", __FUNCTION__, unitPrinter::Inst()->sizeFromB ( size ) );
    exit(EXIT_FAILURE);
  }
}

void getNextFileNmae(char* filename, char *basename)
{
  char dir[1024];
  char fname[1024];
  char ext[1024];
  
  splitFileName(basename, dir, fname, ext);
  
  int i = 0;
  
  std::ifstream* my_file = NULL;
  
  do
  {
    if ( dir[0])
      sprintf ( filename, "%s/", dir);
    else
      strClear(filename);
    
    sprintf ( filename, "%s%s-%05i", filename, fname, i);
    
    if ( ext[0])
      sprintf ( filename, "%s.%s", filename, ext);

    i++;
    
    if ( my_file )
      delete my_file;
    
    my_file = new std::ifstream ( filename );
    
    if ( i >= 100000 )
      return;
    
  }
  while ( my_file->good() );
  
  my_file->close();
}

int compareFloat (const void * a, const void * b)
{
  if ( *(float*)a <  *(float*)b ) return -1;
  if ( *(float*)a == *(float*)b ) return 0;
  if ( *(float*)a >  *(float*)b ) return 1;
};





