#ifndef UTIL_H_INCLUDED
#define UTIL_H_INCLUDED

#include <math.h>
#include <iostream>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "utilstats.h"
#include "utiltime.h"
#include "utilTimer.h"

#define FOLD if(1)
#define FOUT if(0)

// Free a pointer and set value to zero
#define NULLIFY(pointer) { if (pointer) delete ( pointer ); pointer = NULL; }

// Compare and set
#define MAXX(a, b)  ( a=(a)<(b)?b:a )

#define MAX(a,b)  MAX2(a,b)
#define MAX2(a, b) ( ((a)<(b))?(b):(a) )
#define MAX3(a, b, c) MAX2( MAX2(a, b), c )
#define MAX4(a, b, c, d) MAX2( MAX2(a, b), MAX2(c, d) )
#define MAX5(a, b, c, d, e) MAX3( MAX2(a, b), MAX2(c, d), e )
#define MAX6(a, b, c, d, e, f) MAX3( MAX2(a, b), MAX2(c, d), MAX2(e, f) )
#define MAX7(a, b, c, d, e, f, g) MAX3( MAX2(a, b), MAX2(c, d), MAX2(e, f), g )

// Compare pare and set
#define MINN(a, b)  (a=(a)>(b)?(b):(a) )

#define MIN(a,b)  MIN2(a,b)
#define MIN2(a, b) ( ((a)>(b))?(b):(a) )
#define MIN3(a, b, c) MIN2( MIN2(a, b), c )
#define MIN4(a, b, c, d) MIN2( MIN2(a, b), MIN2(c, d) )
#define MIN5(a, b, c, d, e) MIN3( MIN2(a, b), MIN2(c, d), e )
#define MIN6(a, b, c, d, e, f) MIN3( MIN2(a, b), MIN2(c, d), MIN2(e, f) )
#define MIN7(a, b, c, d, e, f, g) MIN3( MIN2(a, b), MIN2(c, d), MIN2(e, f), g )

// Sway to values with the same data type 
#define xorSwap(x,y) x ^= y; y ^= x; x ^= y;

//Simple square
#define SQR(x) ((x)*(x))

// A macro to do safe division ( checks for a 0 denominator)
#define fDivide(numerator, denominator) (float)denominator==0.0f?0.0f:numerator/(float)denominator

// Clear a c string
#define strClear(string) string[0]=0

// A macro to do safe division ( checks for a 0 denominator)
#define dDivide(numerator, denominator) (double)denominator==0.0f?0.0f:numerator/(double)denominator

#define TMP { int tmp = 0; };
#define BREAK int tmp = 0;

#define MAXLINE 1024

//#define INT      1 << 1
//#define FLOAT    1 << 2
//#define DOUBLE   1 << 3
//#define STRING   1 << 4

#define BLACK     "\033[22;30m"
#define GREEN     "\033[22;31m"
#define MAGENTA   "\033[22;35m"
#define RESET     "\033[0m"

// Percentage ( 0 - 1.0 ) between two values
#define PER(value, min, max) ( (value-min)/(float)(max-min) )
#define PERD(value, min, max) ( (value-min)/(double)(max-min) )

#define UNIT unitPrinter::Inst()

const double DINF = 1.0/0.0;

/** Detect if running in GDB
 * Taken from http://stackoverflow.com/questions/3596781/detect-if-gdb-is-running
 * gdb apparently opens FD(s) 3,4,5 (whereas a typical prog uses only stdin=0, stdout=1,stderr=2)
 */
int detect_gdb(void);

inline int midpointI(int i1, int i2)
{
  return floor ( (i1+i2)/2.0 ) ;
}

/**
 * @brief function call back that accepts a single pointer.
 **/
typedef void* (*pointerCallback)(void*);

/**
 * @brief A structure to hold version numbers for library's
 **/
struct versionNumber
{
  int major;
  int minor;
  int build;
};

/**
 * @brief A structure to hold a floating point complex  number
 **/
struct complexNumf
{
  float r;
  float i;
  
  // default + parametrized constructor
  complexNumf(float r=0.0, float i=0.0) 
  : r(r), i(i)
  {
  }
  
  // assignment operator
  complexNumf& operator=(const complexNumf& a)
  {
    r=a.r;
    i=a.i;
    return *this;
  }
  
  // assignment operator
  complexNumf& operator=(const float a)
  {
    r=a;
    i=a;
    return *this;
  }
  
  // assignment operator
  complexNumf& operator=(const double a)
  {
    r=a;
    i=a;
    return *this;
  }
  
  // assignment operator
  complexNumf& operator=(const int a)
  {
    r=a;
    i=a;
    return *this;
  }
  
  // assignment operator
  bool operator==(const complexNumf& a)
  {
    if (r == a.r && i == a.i )
      return true;
    else
      return false;
  }
};

/**
 * @brief A class to convert unit to human readable outputs
 **/
class unitPrinter
{
  public:
    static unitPrinter* Inst();
    static unitPrinter* Instance() ;

    char *sizeFromB ( double byte );
    char *timeFromS ( double seconds );
    char *timeFromMS ( double milliSeconds );

  private:
    
    unitPrinter() {}; // Private so that it can  not be called
    unitPrinter ( unitPrinter const& ) {};         // copy constructor is private

    static unitPrinter* m_pInstance;

    static const int len = 1024;
    char string[len];
};

double calcDist ( int noDims, float* v1, float* v2 );
double calcDist ( int noDims, int* v1, int* v2 );

/**
 * @brief get the number of processors 
 * @return the number of processors
 **/
int getNoProc();

void getCpuName ( char* name, int core = 0 );

/**
 * @brief get total system RAM in bytes
 *
 * @return number of bytes of total system RAM
 **/
ulong getTotalSystemMemory();

/**
 * @brief get total system RAM in bytes
 *
 * @return number of bytes of total system RAM
 **/
ulong getTotalRam();

/**
 * @brief get free ram in bytes
 *
 * @return number of bytes of free RAM
 **/
ulong getFreeRam();

/**
 * @brief decompose a full path into is parts
 *
 * assumes directory, file and extension are initialised, if they are NULL no value will be written to them
 * 
 * @param fName the file to be decomposed
 * @param directory the directory will be written here (everything before the last slash)
 * @param file      the file name will be written here (everything between the last slash and the last period)
 * @param extension the file extension will be written here (everything after the last period)
 * @return a pointer to the beginning if the file name original file name 
 **/
char* splitFileName(char *fName, char* directory, char *file, char *extension); 

/**
 * @brief printf a message iff compiled in debug mode
 *
 * @param format C string that contains a format string that follows the same specifications as format in <a href="http://www.cplusplus.com/printf">printf</a>
 * @return void
 **/
inline void debugMSG ( const char* format, ... )
{
#ifdef DEBUG
  if ( detect_gdb() )
  {
    //printf("in GDB\n");
    va_list ap;
    va_start ( ap, format );
    vprintf ( format, ap );      // Write the line
    va_end ( ap );

    std::cout.flush();
  }
  else
  {
    //printf("NOT in GDB\n");
    printf ( MAGENTA );

    va_list ap;
    va_start ( ap, format );
    vprintf ( format, ap );      // Write the line
    va_end ( ap );

    printf ( RESET );
    std::cout.flush();
  }


#endif
}

/**
 * @brief replaced all occurrences of on char with another in a string
 *
 * @param str      the string
 * @param replace  the char to search for
 * @param ch       the car to replace it with
 * @return void
 **/
void replaceChar(char* str, const char replace, const char ch);

/**
 * @brief Set the number of threads open MP will use
 * 
 * If the integer passed contains 0, the thread count will be set to the processor count (this is a standard practice)
 * the integer thread count will the be set to this value.
 *
 * If the pointer is set to NULL, the thread count will be set to the processor count (this is a standard practice)
 *
 * @param threadCount the desired number of threads if this is set to 0 or NULL processor count will be used.
 * @return the number of threads that were active at the time of this call
 **/
int setOmpThreadCound(int *threadCount = NULL );

int getFileList (char* directory, char* names, char* namesList);

void getNextFileNmae(char* filename, char *basename);

int compareFloat (const void * a, const void * b);

void* safemalloc (ulong size);

#endif // UTIL_H_INCLUDED
