#ifndef TIMER_H
#define TIMER_H

#ifdef _WIN32

#include <Windows.h>
class timer
{
  private:
    unsigned int startTime;

  public:
    timer() { startTime = GetTickCount(); };
    ~timer();

    void Start() { startTime = GetTickCount(); };
    unsigned int Stop() { return GetTickCount() - startTime; };
    unsigned int Reset() { unsigned int tmp = startTime; startTime = GetTickCount(); return GetTickCount() - tmp; };
};

#elif __linux__

#include <sys/time.h>
#include <time.h>

/** a timer class for precision timing
 * This class times to the microsecond level using gettimeofday
 * Or to the nanosecond using the posix clock_gettime
 */
class timer
{
  public:
    /** Constructor
     * The constructor will start the timer
     */
    timer() { Start(); };

    /** Destructor */
    ~timer() {} ;

    /** Start the timer */
    void Start() { gettimeofday ( &start, NULL ); clock_gettime(CLOCK_MONOTONIC, &pstart); };

    /** Stop the timer
     * @return The number of microseconds since the timer was created or last started or reset
     */
    unsigned int Stop() { gettimeofday ( &end, NULL ); return ( ( end.tv_sec  - start.tv_sec ) * 1000 + ( end.tv_usec - start.tv_usec ) / 1000.0 ) + 0.5; };
    
    /** Stop the timer
     * @return The number of seconds since the timer was created or last started or reset
     */
    double stopS() { clock_gettime(CLOCK_MONOTONIC, &pend); return ( ( pend.tv_sec  - pstart.tv_sec ) + (double)( pend.tv_nsec - pstart.tv_nsec )*1e-10); };

    /** Reset the timer
     * @return The number of microseconds since the timer was created or last started or reset
     */
    unsigned int Reset() { unsigned int tmp = Stop(); Start(); return tmp; };

  private:
    struct timeval start, end, timeval;
    struct timespec pstart, pend;
};
#endif

#endif // TIMER_H
